import 'dart:async';
import 'dart:convert';

import 'package:apex_socket/apex_api.dart';
import 'package:apex_socket/cipher/crypto.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:logging/logging.dart';
import 'package:socket_io_client/socket_io_client.dart' as IO;

export 'package:socket_io_client/socket_io_client.dart' show OptionBuilder;

typedef dynamic EventHandler<T>(T data);

class ApexSocket extends ChangeNotifier {
  final log = Logger('ApexSocket');

  IO.Socket socket;

  /// Check if socket object is created and connected as well :)
  bool get isConnected => socket != null && socket.connected;

  final KeyPair _iosKey;
  final KeyPair _androidKey;
  final KeyPair _webKey;

  final ApexApiConfig config;

  String get _secretKey => kIsWeb
      ? _webKey.secretKey
      : (defaultTargetPlatform == TargetPlatform.android
          ? _androidKey.secretKey
          : _iosKey.secretKey);

  String get _publicKey => kIsWeb
      ? _webKey.publicKey
      : (defaultTargetPlatform == TargetPlatform.android
          ? _androidKey.publicKey
          : _iosKey.publicKey);

  bool get encrypt => kIsWeb
      ? _webKey != null
      : (defaultTargetPlatform == TargetPlatform.android ? _androidKey != null : _iosKey != null);

  String get _os => defaultTargetPlatform == TargetPlatform.android
      ? 'A'
      : defaultTargetPlatform == TargetPlatform.iOS
          ? 'I'
          : kIsWeb
              ? 'W'
              : 'U';

  ApexSocket(
    this.config, {
    KeyPair iosKey,
    KeyPair androidKey,
    KeyPair webKey,
  })  : _iosKey = iosKey,
        _webKey = webKey,
        _androidKey = androidKey,
        assert(config != null);

  void connect() {
    if (this.socket != null) this.socket.connect();
  }

  void createSocket({
    bool recreate = false,
    EventHandler onConnect,
    EventHandler onConnectError,
    EventHandler onError,
    EventHandler onDisconnect,
    EventHandler onReconnect,
  }) {
    assert(config.options != null, 'Options must not be null');
    if (socket == null || recreate) {
      if (recreate) if (this.socket != null) this.socket.destroy();
      IO.Socket socket = IO.io(
          '${config.url}${config.port != null ? ':${config.port}' : ''}${config.namespace != null ? '/${config.namespace}' : ''}',
          config.options.setTransports(['websocket']).build());

      socket.onConnect((data) {
        notifyListeners();
        if (onConnect != null) return onConnect(data);
      });

      socket.onConnectError(onConnectError);

      socket.onError(onError);

      socket.onDisconnect((data) {
        notifyListeners();
        if (onDisconnect != null) return (data);
      });

      socket.onReconnect(onReconnect);

      this.socket = socket;
    }

    if (!this.socket.connected)
      this.socket.connect();
    else
      log.info('already connected');
  }

  String _encryptedData(Crypto crypto, String data, bool enc) {
    String data = '';
    bool shouldEncrypt = ((enc == null) ? encrypt : enc);

    if (shouldEncrypt) {
      data = crypto.encrypt(data);
    } else {
      data = base64Encode(utf8.encode(data));
    }

    return data;
  }

  String _decryptedData(Crypto crypto, String data, bool enc) {
    bool shouldEncrypt = ((enc == null) ? encrypt : enc);

    if (shouldEncrypt) {
      try {
        return crypto.decrypt(data);
      } catch (ignore) {
        throw DecryptionError('Could not decrypt!');
      }
    }
    return data;
  }

  void privateEmit({String join, String data, bool enc}) {
    this.socket.emit(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 1,
          'version': config.privateVersion,
          'data': _encryptedData(Crypto(_secretKey, _publicKey), data, enc),
          'join': join,
        }));
  }

  void publicEmit({String join, String data}) {
    data = base64Encode(utf8.encode(data));
    this.socket.emit(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 0,
          'version': config.publicVersion,
          'data': data,
          'join': join,
        }));
  }

  void privateEmitWithAck(String data, {String join, Function(String data) ack, bool enc}) {
    if (config.mockEnabled && config.mocks.any((element) => element.requestData == data)) {
      ack(config.mocks.firstWhere((element) => element.requestData == data).responseData);
      return;
    }

    final crypto = Crypto(_secretKey, _publicKey);
    this.socket.emitWithAck(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 1,
          'version': config.privateVersion,
          'data': _encryptedData(crypto, data, enc),
          'join': join,
        }), ack: (m) {
      ack(_decryptedData(crypto, m, enc));
    });
  }

  void publicEmitWithAck(String data, {String join, Function(String data) ack, bool enc}) {
    if (config.mockEnabled && config.mocks.any((element) => element.requestData == data)) {
      ack(config.mocks.firstWhere((element) => element.requestData == data).responseData);
      return;
    }

    final crypto = Crypto(_secretKey, _publicKey);

    this.socket.emitWithAck(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 0,
          'version': config.publicVersion,
          'data': data,
          'join': join,
        }), ack: (m) {
      ack(_decryptedData(crypto, m, enc));
    });
  }

  Future privateEmitWithCompleter(String data, {String join, bool enc}) {
    Completer completer = Completer();

    if (config.mockEnabled && config.mocks.any((element) => element.requestData == data)) {
      if (!completer.isCompleted)
        completer.complete(
            config.mocks.firstWhere((element) => element.requestData == data).responseData);
      return completer.future;
    }

    final crypto = Crypto(_secretKey, _publicKey);
    _encryptedData(crypto, data, enc);

    this.socket.emitWithAck(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 1,
          'version': config.privateVersion,
          'data': data,
          'join': join,
        }), ack: (m) {
      try {
        if (!completer.isCompleted) completer.complete(_decryptedData(crypto, m, enc));
      } on DecryptionError {
        if (!completer.isCompleted)
          completer.completeError(DecryptionError('could not decrypt $m'));
      }
    });

    this.socket.on('connect_error', (data) {
      if (!completer.isCompleted) completer.completeError(ConnectError(data.toString()));
    });

    return completer.future;
  }

  Future publicEmitWithCompleter(String data, {String join}) {
    final completer = Completer();

    if (config.mockEnabled && config.mocks.any((element) => element.requestData == data)) {
      if (!completer.isCompleted)
        completer.complete(
            config.mocks.firstWhere((element) => element.requestData == data).responseData);
      return completer.future;
    }

    data = base64Encode(utf8.encode(data));

    this.socket.emitWithAck(
        config.eventName,
        jsonEncode({
          'os': _os,
          'private': 0,
          'version': config.publicVersion,
          'data': data,
          'join': join,
        }), ack: (m) {
      if (!completer.isCompleted) completer.complete(m.toString());
    });

    this.socket.on('connect_error', (data) {
      if (!completer.isCompleted) completer.completeError(ConnectError(data.toString()));
    });

    return completer.future;
  }
}

class NullResponseError implements Exception {
  final String message;

  NullResponseError(this.message);

  @override
  String toString() {
    return this.message;
  }
}

class ConnectError implements Exception {
  final String message;

  ConnectError(this.message);

  @override
  String toString() {
    return this.message;
  }
}

class DecryptionError implements Exception {
  final String message;

  DecryptionError(this.message);

  @override
  String toString() {
    return this.message;
  }
}
